# -*- coding: utf-8 -*-
"""Define box border constants."""


class Border:
    """Box drawing characters, includeing characters for splits."""

    HORIZONTAL = "─"
    VERTICAL = "│"
    TOP_LEFT = "╭"
    TOP_RIGHT = "╮"
    BOTTOM_LEFT = "╰"
    BOTTOM_RIGHT = "╯"
    SPLIT_BOTTOM = "┴"
    SPLIT_TOP = "┬"
    SPLIT_LEFT = "├"
    SPLIT_RIGHT = "┤"
    CROSS = "┼"
