# -*- coding: utf-8 -*-
"""Instantiates the euporie package."""

__app_name__ = "euporie"
__version__ = "0.1.15"
__logo__ = "⚈"
__strapline__ = "A TUI editor for Jupyter notebooks"
__author__ = "Josiah Outram Halstead"
__email__ = "josiah@halstead.email"
__copyright__ = f"© 2021, {__author__}"
__license__ = "MIT"

import euporie.log  # noqa E401
